#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Decoder classes, Controller architecture in a MVC layout."""
from __future__ import absolute_import
import os
from dnazip.sequence import Sequence
from dnazip.burros_wheeler import BurrosWheeler
from dnazip.huffman import HuffmanTree

class HuffDecoder:
    """A decoder class for Huffman decompression, it is used as a controller
    in a MVC architecture.
    
    Attributes
    ----------
    path: str
        The path of the file to be decompressed with Huffman decompression.
    seq: Sequence
        The sequence that was extracted from the file; a Sequence object.
    dehuffman_output: str
        The output file path for Huffman decompression.
    binary: str
        The binary sequence that was translated from unicode using Huffman
        codes.
    header: str
        The header of the compressed file; contains Huffman codes and paths
        as well as padding that were generated when compressing the sequence.
    unicode: str
        The compressed format of the sequence.
    decompressed: str
        The decompressed sequence; normally the Burros-Wheeler transform of an
        original sequence.
    """

    def __init__(self: object, path: str) -> None:
        """Class constructor.

        Parameters
        ----------
        path : str
            The path of the file to be read.

        Returns
        -------
        None
            A class instance.

        """
        self.path = os.path.splitext(path)[0]
        self.seq = Sequence(path)
        self.dehuffman_output = self.path + '_dehuff.txt'
        self.binary = None
        self.header = None
        self.unicode = None
        self.decompressed = None

    def decode(self: object) -> None:
        """The main decoding method of the controller.

        Returns
        -------
        None
            Fills all the properties of an object and writes out the
            decompressed sequence to a file.

        """
        seq = self.seq.read_bytes()
        self.header = seq[:seq.index('\n')]
        self.unicode = seq[seq.index('\n')+1:]
        re_codes = HuffmanTree.header_to_codes(self.header)
        binary = HuffmanTree.unicode_to_binstr(self.unicode)
        padding = int(re_codes['pad'])
        self.binary = HuffmanTree.remove_padding(binary, padding)
        self.decompressed = HuffmanTree.binstr_to_seq(self.binary, re_codes)
        Sequence(self.dehuffman_output).write(self.decompressed)

class BWDecoder:
    """A decoder class for inversing the Burros-Wheeler transform, it is used 
    as a controller in a MVC architecture.
    
    Attributes
    ----------
    path: str
        The path of the file to be transformed with inverse of Burros-Wheeler.
    seq: Sequence
        The sequence that was extracted from the file; a Sequence object.
    debwt_output: str
        The output file path for the inverse of BWT.
    bwm: List[str]
        The reconstructed Burros-Wheeler Matrix.
    original: str
        The original sequence from the inverse BWT.
    """
    def __init__(self: object, path: str) -> None:
        """Class constructor.

        Parameters
        ----------
        path : str
            The path of the file to be read.

        Returns
        -------
        None
            A class instance.

        """
        self.path = os.path.splitext(path)[0]
        self.seq = Sequence(path)
        self.debwt_output = self.path + '_debwt.txt'
        self.bwm = None
        self.original = None
        
    def decode(self: object) -> None:
        """The main decoding method of the controller.

        Returns
        -------
        None
            Fills all the properties of an object and writes out the original
            sequence to a file.

        """
        self.bwm = list(BurrosWheeler.reconstruct_bwm(self.seq.read()))
        self.original = BurrosWheeler.decode_bwt(self.bwm[-1])
        Sequence(self.debwt_output).write(self.original)

class FullDecoder:
    """A decoder class for both Huffman decompression and the inverse of the
    Burros-Wheeler transform, controller architecture.

    Attributes
    ----------
    path: str
        The path of the file to be compressed with BWT + Huffman compression.
    huff_decoder: HuffDecoder
        A HuffDecoder object to do the Huffman decompression on a compressed
        sequence.
    bw_decoder: BWDecoder
        A BWDecoder object to do the inverse of Burros-Wheeler transform on
        the output of Huffman decompression.
    """

    def __init__(self: object, path: str) -> None:
        """Class constructor.

        Parameters
        ----------
        path : str
            The path of the file to be read.

        Returns
        -------
        None
            A class instance.

        """
        self.path = path
        self.huff_decoder = None
        self.bw_decoder = None
        
    def full_unzip(self: object) -> None:
        """The main decoding method of the controller, it first decodes the
        sequence with Huffman decompression, then passes the uncompressed
        sequence to do the Inverse of Burros-Wheeler transform and obtain
        the original sequence.

        Returns
        -------
        None
            Fills all the properties of an object and writes out the
            original sequence to a file.

        """
        self.huff_decoder = HuffDecoder(self.path)
        self.huff_decoder.decode()

        self.bw_decoder = BWDecoder(self.huff_decoder.dehuffman_output)
        self.bw_decoder.decode()



# -*- coding: utf-8 -*-
"""
Burros-Wheeler Algorithm Class, contains naive and advanced construction
methods, static methods were written as explicit as possible and were
factorized to facilitate algorithmic readability.
The yielding in some functions doesn't respect Space complexity, but it was done in this
manner to facilitate the data flow into the View class (the GUI).
"""
from __future__ import absolute_import
from typing import List, Tuple

class BurrosWheeler:
    """A class to represent the Burros-Wheeler algorithm, all methods are
    static for ease of reading and outside usability.
    """
    @staticmethod
    def pprint(mat: List[str]) -> None:
        """Pretty print, this method prints a burros wheeler matrix
        beautifully (without lists and strings).

        Parameters
        ----------
        mat : List[str]
            The Burros-Wheeler matrix, i.e; a list of strings.

        Returns
        -------
        None
            Prints the matrix.
        """
        for line in mat:
            print(*line, sep="") # scatter operator to print all elements
                                 # of a line

    @staticmethod
    def string_rotations(seq: str) -> List[str]:
        """Returns all string rotations of a sequence.

        Parameters
        ----------
        seq : str
            he sequence to be rotated.

        Returns
        -------
        List[str]
            Returns a list of strings, i.e; a BW matrix like object.

        """

        seq += '$'
        double_seq = seq * 2
        all_rotations = []

        for i in range(0, len(seq), 1):

            rot = double_seq[i:i+len(seq)]
            all_rotations.append(rot)

            yield [rot for rot in all_rotations]

    @staticmethod
    def construct_bwm(rotations: List[str]) -> List[str]:
        """This method constructs the Burros-Wheeler Matrix from a list of
        string rotations.

        Parameters
        ----------
        rotations : List[str]
            A list of strings, i.e; a BW matrix like object.

        Returns
        -------
        List[str]
            A list of strings or a Burros-Wheeler Matrix.

        """

        sorted_rotations = sorted(rotations)
        return sorted_rotations

    @staticmethod
    def encode_bwt(matrix: List[str]) -> str:
        """Returns the Burros-Wheeler Transform from a given Burros-Wheeler
        Matrix. the Burros-Wheeler Transform corresponds to the last column
        of the matrix.

        Parameters
        ----------
        matrix : List[str]
            A Burros-Wheeler Matrix.

        Returns
        -------
        str
            The Burros-Wheeler Transform.

        """

        last_column = []
        for line in matrix:
            last_char = line[-1]
            last_column.append(last_char)

        transformed_seq = ''.join(last_column)
        return transformed_seq

    @staticmethod
    def reconstruct_bwm(bwt: str) -> List[str]:
        """This method reconstructs the Burros-Wheeler Matrix given the
        corresponding Burros-Wheeler Transform. The naive algorithm for
        constructing the matrix given the transform is going to iteratively
        add the transform as a left column, then sorts lexicographically
        the columns.

        Parameters
        ----------
        bwt : str
            The Burros-Wheeler Transform.

        Returns
        -------
        List[str]
            A Burros-Wheeler Matrix.

        """
        bwm = []
        # first loop to create seeds for lines O(n)
        for _ in range(0, len(bwt), 1):
            bwm.append('')

        for _ in range(0, len(bwt), 1):

            for i in range(0, len(bwt), 1):
                bwm[i] = bwt[i] + bwm[i]
 
            yield [line for line in bwm]
            bwm.sort()
            yield [line for line in bwm]

    @staticmethod
    def decode_bwt(matrix: List[str]) -> str:
        """This method returns the original sequence from a given
        Burros-Wheeler Matrix, the original sequence is the line that ends
        with the character '$'.

        Parameters
        ----------
        matrix : List[str]
            A Burros-Wheeler Matrix.

        Returns
        -------
        str
            The original sequence.

        """

        seq = ""
        for line in matrix: # search for the line that ends with '$'
            if line[-1] == "$":
                seq += line

        return seq[:-1] # return the sequence without the '$' sign

    @staticmethod
    def suffix_array(sequence: str) -> List[Tuple[str, int]]:
        """Builds a suffix-array from a given sequence of characters.
        - Complexity of the algorithm O(n^2log(n))
        - Sorting is O(nlogn) and Finding is O(n)

        Parameters
        ----------
        sequence : str
            The given sequence of characters.

        Returns
        -------
        List[Tuple[str, int]]
            The suffix array of the sequence; a list of tuples.

        """

        sequence += '$'
        suff_arr = []
        for i in range(0, len(sequence), 1):
            suff_arr.append((sequence[i:], i))

        return sorted(suff_arr)

    @staticmethod
    def bwt_advanced(sequence: str) -> str:
        """Generates a Burros-Wheeler Transfrom from a suffix array, advanced
        construction of BWT. Better algorithmic complexity.

        Parameters
        ----------
        sequence : str
            The sequence to ve transformed.

        Returns
        -------
        str
            The Burros-Wheeler Transform.

        """

        bwt = []
        for suff in BurrosWheeler.suffix_array(sequence):
            i = suff[1] # The suffix's index is the 2nd element in the tuple
            if i == 0:
                bwt.append('$')
            else:
                bwt.append(sequence[i - 1])

        return ''.join(bwt)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Main script to launch the GUI."""
from __future__ import absolute_import
from dnazip.interface import Interface

def main() -> None:
    """Launches the GUI."""
    GUI = Interface()
    GUI.main()

if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
"""Huffman coding algorithm classes, contains node and tree separate classes."""
from __future__ import absolute_import
from typing import Dict

class HuffmanNode:
    """A class to represent heap nodes of a huffman coding tree.

    Attributes
    ----------
    char: str
        The character to be saved as a node value.
    freq: int
        The frequency of the char.
    right_child: HuffmanNode
        The right child of the heap node.
    left_child: HuffmanNode
        The left child of the heap node.
    dir: str
        The path of the current node in the Huffman tree, usually a string of
        0s and 1s.
    """
    def __init__(self: object, char: str, freq: int, left: object=None, right: object=None) -> None:
        """Class constructor

        Parameters
        ----------
        char : str
            A character (Huffman coding characters).
        freq : int
            The frequency of the character.
        left : object, optional
            The left child of the heap node. The default is None.
        right : object, optional
            The right child of the heap node. The default is None.

        Returns
        -------
        None
            A class instance.

        """
        self.char = char
        self.freq = freq
        self.__left_child = left
        self.__right_child = right
        self.dir = ''

    @property
    def right_child(self: object) -> object:
        """The getter of the right child.

        Returns
        -------
        object

        """
        return self.__right_child

    @right_child.setter
    def right_child(self: object, right_child: object) -> None:
        """The setter of the right child.

        Parameters
        ----------
        right_child : object
            The new child.

        Returns
        -------
        None

        """
        self.__right_child = right_child

    @property
    def left_child(self: object) -> object:
        """The getter of the left child.

        Returns
        -------
        object

        """
        return self.__left_child

    @left_child.setter
    def left_child(self: object, left_child: object) -> None:
        """The setter of the left child.

        Parameters
        ----------
        right_child : object
            The new child.

        Returns
        -------
        None

        """
        self.__left_child = left_child

    def __str__(self: object) -> str:
        """Returns a string representation using print().

        Returns
        -------
        str
            A string to be displayed with print().

        """
        return "Hello! I'm a Huffman coding HeapNode class instance."

    def __repr__(self: object) -> str:
        """A coder friendly representation of the HuffmanNode object.

        Returns
        -------
        str
            A string.

        """
        return "Node(%s, freq=%s, right=%s, left=%s)"%(self.char,
                                                       self.freq,
                                                       self.right_child,
                                                       self.left_child)

class HuffmanTree:
    """A class to represent a huffman tree.

    Attributes
    ----------
    sequence: str
        The sequence to be encoded, i.e; the Burros-Wheeler transform of the
        sequence.
    frequency: Dict[str, int]
        A dictionary to store frequency values for each character of the
        sequence, the character as a key and its frequency as a value.
    root: HuffmanNode
        The root node of the Huffman tree.
    codes: Dict[str, str]
        The Huffman codes for a given sequence, The character as a key and
        its tree path as a value(a string).
    pad: int
        The padding of the sequence, i.e; the number of zeroes that was added
        to the end of sequence until it was divisible by 8 (coding in 8 bits).

    """
    def __init__(self: object, sequence: str) -> None:
        """The class constructor.

        Parameters
        ----------
        sequence : str
            The sequence to be coded.

        Returns
        -------
        None
            A class instance.

        """
        self.sequence = sequence
        self.frequency = HuffmanTree.freq_dict(self.sequence)
        self.root = self.create_tree()
        self.codes = {}

    @staticmethod
    def freq_dict(sequence: str) -> Dict[str, int]:
        """This method returns the frequence dictionary of a given sequence.

        Parameters
        ----------
        sequence : str
            The sequence of interest.

        Returns
        -------
        Dict[str, int]
            A dictionary of character frequencies.

        """
        f_dict = {}
        for char in sequence:
            if char in f_dict.keys():
                f_dict[char] += 1
            else:
                f_dict[char] = 1
        return f_dict

    def create_tree(self: object) -> HuffmanNode:
        """The main algorithm for creating The Huffman tree.

        Returns
        -------
        HuffmanNode
            The root node of the tree.

        """
        leafs = []
        for char, freq in self.frequency.items():
            leafs.append(HuffmanNode(char, freq))

        while len(leafs) > 1:

            leafs = sorted(leafs, key=lambda x: x.freq)

            left = leafs.pop(0)
            right = leafs.pop(0)

            new_char = left.char + right.char
            new_freq = left.freq + right.freq

            left.dir = "0"
            right.dir = "1"

            new_node = HuffmanNode(new_char, new_freq, left, right)
            leafs.append(new_node)

        return leafs[0]

    def get_codings(self: object, node: HuffmanNode, val: str='') -> None:
        """A depth-first search recursive algorithm to find the paths of each
        character in the tree.

        Parameters
        ----------
        node : HuffmanNode
            The starting node, the root of the tree.
        val : str, optional
            The path of the current recursion in a tree. The default is ''.

        Returns
        -------
        None
            Fills the codes dictionary property of the Huffman Tree.

        """

        curr_path = val + node.dir

        if node.left_child:
            self.get_codings(node.left_child, curr_path)
        if node.right_child:
            self.get_codings(node.right_child, curr_path)

        if not node.left_child and not node.right_child:
            self.codes[node.char] = curr_path

    def seq_to_binstr(self: object) -> str:
        """This method transforms the current sequence of the Huffman tree
        to a binary sequence using the codes (paths of every character), it
        also adds a padding to the end of the binary sequence (a variable
        number of zeroes, when sequence is divisible by 8 then add 8 zeroes
        by default). The binary sequence will be coded in 8-bits later.
        It saves the padding in the codes property.

        Returns
        -------
        str
            The binary sequence.

        """

        bin_str = ""
        for char in self.sequence:
            bin_str += self.codes[char]

        pad = 8 - len(bin_str) % 8
        if pad != 0:
            for _ in range(0, pad, 1):
                bin_str += '0'

        self.codes['pad'] = str(pad) #Save the padding value to codes

        return bin_str

    @staticmethod
    def binstr_to_unicode(bin_str: str) -> str:
        """This method codes a binary sequence in 8-bits in UTF-8.

        Parameters
        ----------
        bin_str : str
            The binary sequence to be coded.

        Returns
        -------
        str
            The unicode string that corrsponds to the binary sequence.

        """

        unicode = ""
        for bit in range(0, len(bin_str), 8):
            eight_bits = bin_str[bit:bit+8]
            code = int(eight_bits, 2)
            unicode += chr(code)

        return unicode

    @staticmethod
    def unicode_to_binstr(unicode: str) -> str:
        """Transforms a unicode sequence to a binary sequence.

        Parameters
        ----------
        unicode : str
            The unicode sequence to be transformed.

        Returns
        -------
        str
            The corrsponding binary string.

        """

        bin_str = ""
        for uni in unicode:
            code = ord(uni)
            bin_str += '{:08b}'.format(code)

        return bin_str

    @staticmethod
    def remove_padding(bin_str: str, pad: int) -> str:
        """This function removes the padding from a given binary sequence,
        the padding is normally a variable number of zeroes added when coding
        in 8-bits.

        Parameters
        ----------
        bin_str : str
            The binary string to be stripped from a sequence of zeroes.
        pad : int
            The padding, the number of zeroes at the end of the binary
            sequence.

        Returns
        -------
        str
            The no-padded binary string.

        """
        return bin_str[:-pad]

    @staticmethod
    def binstr_to_seq(bin_str: str, codes: Dict[str, str]) -> str:
        """Transforms a binary string to a sequence given a codes dictionary
        of paths.

        Parameters
        ----------
        bin_str : str
            The binary string to be transformed.
        codes : Dict[str, str]
            A dictionary of characters alongside their paths.

        Returns
        -------
        str
            The original sequence.

        """

        original_seq = ""
        reading_stream = ""
        for num in bin_str:
            reading_stream += num
            for char, path in codes.items():
                if path == reading_stream:
                    original_seq += char
                    reading_stream = ""
                    break

        return original_seq

    def codes_to_header(self: object) -> str:
        """This method transforms the codes of a given Huffman tree object
        into a header to be saved in the compressed file (useful for
        when decompressing later).

        Returns
        -------
        str
            The header of the compressed file.

        """

        header = ""
        for char, path in self.codes.items():
            header += char + "," + path + ";"

        return header + "\n"

    @staticmethod
    def header_to_codes(header: str) -> Dict[str, str]:
        """This method takes the header of a compressed file and transforms
        it into a codes dictionary that's useful when decompressing a sequence.

        Parameters
        ----------
        header : str
            The header of the compressed file.

        Returns
        -------
        Dict[str, str]
            A codes dictionary which maps from characters to thier given paths.

        """
        reconstructed_codes = {}
        for code in header.split(";")[:-1]:
            char, path = code.split(",")
            reconstructed_codes[char] = path
        return reconstructed_codes



#!/usr/bin/env python3
# -*- coding: utf-8 -*
"""View architecture of the main application, i.e; a GUI."""
from __future__ import absolute_import
import os
from tkinter import Tk, Toplevel, filedialog, Menu, messagebox, ttk
from tkinter import Label, Entry, Button, StringVar, Text, Scrollbar
from tkinter import RIGHT, Y, X,  HORIZONTAL, BOTTOM, NONE, END
from typing import Iterator, Generator
from dnazip.sequence import Sequence
from dnazip.decoder import HuffDecoder, BWDecoder, FullDecoder
from dnazip.encoder import HuffEncoder, BWEncoder, FullEncoder

class Interface(Tk):
    """View class of the application using a Tkinter interface."""

    def __init__(self: object) -> None:
        """Class constructor

        Returns
        -------
        None
            An interface instance.

        """
        super().__init__()
        self.title("dnazip")
        self.configure(bg='#ebebeb')
        self.create_main()
        self.create_menu()
        self.create_buttons()
        self.file = None

    def create_main(self: object) -> None:
        """This method is used to create the main menu interface of the GUI."""

        welcome_msg = "Welcome to dnazip! \n A graphical " + \
            "representation of Burros-Wheeler and Huffman " + \
            "Coding algorithms"
        rand = "You don't have a DNA sequence? we'll generate a small " + \
            "random one for you"
        Label(self, text=welcome_msg, bg='#ebebeb', font=(None, 15)).grid(row=0, columnspan=3, padx=5, pady=6)
        Label(self, text="Compression", bg='#ebebeb', borderwidth=2, relief="groove", font=(None, 13, 'bold')).grid(row=1, columnspan=3, padx=5, pady=6)
        Label(self, text="Decompression", bg='#ebebeb', borderwidth=2, relief="groove", font=(None, 13, 'bold')).grid(row=5, columnspan=3, padx=5, pady=6)
        separator_one = ttk.Separator(self, orient='horizontal')
        separator_one.grid(row=4, columnspan=3, ipadx=500, padx=5, pady=6) 
        Label(self, text=rand, bg='#ebebeb', font=(None, 12, 'bold')).grid(row=8, columnspan=3, padx=5, pady=6)
        self.random = StringVar()
        self.entry = Entry(self, textvariable=self.random, width=50)
        self.entry.grid(row=9, columnspan=3, padx=5, pady=10)
        separator_two = ttk.Separator(self, orient='horizontal')
        separator_two.grid(row=7, columnspan=3, ipadx=500, padx=5, pady=6) 

    def create_buttons(self: object) -> None:
        """This method creates buttons for the interface."""
        buttons = {"BWT": Button(
            self, text="BWT", command=self.bwt_window, width = 15).grid(row=3,column=0, padx=5, pady=6),
                   "DEBWT": Button(
                       self, text="reverse BWT", command=self.debwt_window,width = 15).grid(
                           row=6,column=0, padx=5, pady=6),
                   "Huffcode": Button(
                       self, text="Huffman coding", command=self.huffcode_window, width = 15).grid(
                           row=3,column=1, padx=5, pady=6),
                   "Huffdecode": Button(
                       self, text="Huffman decoding", command=self.huffdecode_window, width = 15).grid(
                           row=6,column=1, padx=5, pady=6),
                   "fullzip": Button(
                       self, text="Full zip", command=self.fullzip_window, width = 15).grid(
                           row=3,column=2, padx=5, pady=6),
                   "fullunzip": Button(
                       self, text="Full unzip", command=self.fullunzip_window, width = 15).grid(
                           row=6,column=2, padx=5, pady=6),
                    "generate": Button(
                       self, text="Generate", command=self.generate_random, width = 15).grid(
                           row=10,column=1, padx=5, pady=6),
                     "save": Button(
                       self, text="Save", command=self.save_random, width = 15).grid(
                           row=11,column=1, padx=5, pady=6)}

        self.buttons = buttons

    def open_file(self: object) -> None:
        """This method is used to create a file chooser for the interface."""
        self.file = filedialog.askopenfilename(
            initialdir= os.getcwd(),title="Select File",filetypes=(
                ("Text Files", "*.txt"),("all files","*.*")))

        if self.file:
            messagebox.showinfo("Selected file", "You have selected %s"%(
                self.file))

    def save_random(self: object) -> None:
        """This method is used to save a randomly generated DNA sequence 
        to a file.
        """
        seq_to_save = self.entry.get()
        if seq_to_save:
            path = filedialog.asksaveasfilename(
                initialdir= os.getcwd(),title="Select File", filetypes=(
                    ("Text Files", "*.txt"), ("all files","*.*")))
            if path:
                Sequence(path).write(seq_to_save)
                messagebox.showinfo("Sequence saved", "The sequence was saved to %s" %(
                    path))
        else:
            messagebox.showerror("No sequence entered", "The sequence box is empty")

    def about(self: object) -> None:
        """This method is used to generate the about menu page information."""
        msg = 'This project is developed for the "Sequence Algorithms" ' + \
            'Course as a part of The M.Sc. in Bioinformatics degree at ' + \
                'Aix-Marseille University, France. \n' + \
        'Source code available at: https://github.com/dabane-ghassan/dnazip'
        messagebox.showinfo("About dnazip", msg)

    def generate_random(self: object) -> None:
        """This method is used to generate random sequences of DNA of length
        50 and show them inside the random text box tkinter entry.
        """
        self.random.set(Sequence.generate(length=50))

    def next_btn(self: object, controller: Iterator[str]) -> str:
        """This method is used to create the output of a universal next button
        for all protocols of the program, it will be passed to a given label
        that changes its content.

        Parameters
        ----------
        controller : Iterator[str]
            The controller of a given protocol that contains all steps to show.

        Returns
        -------
        str
            Every step to be shown, one at a time.
        """
        try:
            to_print = next(controller)
            if isinstance(to_print, list):
                return '\n'.join(to_print)
            else:  
                return to_print

        except StopIteration:
            return "The protocole is finished"

    def final_btn(self: object, controller: Iterator[str]) -> str:
        """This method is used to create the output of a universal final step 
        button for all protocols of the program, it will be passed to a given
        label that changes its content.

        Parameters
        ----------
        controller : Iterator[str]
            The controller of a given protocol that contains all steps to show.

        Returns
        -------
        str
            The final step to be shown.
        """
        
        try:
            to_print = list(controller)[-1] # Getting the final result of a generator
            return to_print

        except IndexError:
            return "The protocole is finished, please refer to the main menu."

    def step_by_step(self: object, window: Tk, protocol: Iterator[str], names: Generator)-> None:
        """This method creates a universal step by step advancing interface
        for a given chosen protocole.

        Parameters
        ----------
        window : Tk
            The given window.
        protocol : Iterator[str]
            The output protocol to be displayed.
        names : Generator
            The names of the steps to be displayed.

        """ 
        steps = StringVar()
        current_step = Label(window, textvariable=steps, bg="#ebebeb",
                     font=(None, 9, "bold"), borderwidth=2)
        current_step.place(relx = 0.5, 
                   rely = 0.1, anchor="center")
        xscrollbar = Scrollbar(window, orient=HORIZONTAL)
        xscrollbar.pack(side=BOTTOM, fill=X)

        # Vertical (y) Scroll Bar
        yscrollbar = Scrollbar(window)
        yscrollbar.pack(side=RIGHT, fill=Y)

        # Text Widget
        text = Text(window, wrap=NONE, width=100, height=40,
                    xscrollcommand=xscrollbar.set,
                    yscrollcommand=yscrollbar.set)
        text.place(relx = 0.5,
                   rely = 0.5,
                   anchor = 'center')

        # Configure the scrollbars
        xscrollbar.config(command=text.xview)
        yscrollbar.config(command=text.yview)

        Button(window, text="Next",
               command=lambda : [self.update_text(text,
                   self.next_btn(protocol)),steps.set(
                       self.next_btn(names))]).pack(side="bottom")

        Button(window, text="Final step",
        command=lambda : [self.update_text(text,
            self.final_btn(protocol)), steps.set(
                       self.final_btn(names))]).pack(side="bottom")

    def update_text(self: object, widget: Text, new_text: str) -> None:
        """This method updates the Text tkinter widget with every step
        in every protocol. it deletes its contents and then replaces it with
        new_text parameter.

        Parameters
        ----------
        widget : Text
            Tkinter's Text widget.
        new_text : str
            The new text to be shown in the widget.

        Returns
        -------
        None
            Replaces text in a given Text widget.

        """
        widget.delete("1.0", END)   #Clear the text window so we can write.
        widget.insert(END,new_text) 

    def BW_output(self: object, controller: BWEncoder) -> Iterator[str]:
        """This method is used to collect all output for the BW encoding.

        Parameters
        ----------
        controller : BWEncoder
            The given controller.

        Yields
        ------
        Iterator[str]
            The content to be shown in the interface.

        """
        yield controller.seq.read()
        for rot in controller.rotations:
            yield rot
        yield controller.bwm
        yield controller.bwt

    def bwt_window(self: object) -> None:
        """This method creates a Tkinter Toplevel window for the step-by-step
        BWT protocol, The output file of the protocol will be shown at the 
        beginning.

        """
        if self.file:
            bwt_window = Toplevel(self)
            bwt_window.title("Burros-Wheeler Transform")
            bwt_window.geometry("1000x1000")
            bwt_window.configure(bg='#ebebeb')
            controller = BWEncoder(self.file)
            controller.encode()
            protocol = self.BW_output(controller)
            prot= list(protocol)

            rots = ["Step 2: Generating all rotations of the sequence" for n in range(len(prot) - 3)]
            
            names = (step for step in ["Step 1 : Visualizing the sequence",
                     *rots,
                     "Step 3: Creating the Burros-Wheeler matrix by sorting all rotations",
                     "Step 4: The Burros-Wheeler transform is the last column of the matrix",
                     "Please refer to the main menu to select another sequence"])
            
            self.step_by_step(bwt_window, iter(prot), names)
            self.program_output(bwt_window, controller.bwt_output)

        else:
            self.no_file_error()

    def DeBW_output(self: object, controller: BWDecoder) -> Iterator[str]:
        """This method is used to collect all output for the BW decoding.

        Parameters
        ----------
        controller : BWDecoder
            The given controller.

        Yields
        ------
        Iterator[str]
            The content to be shown in the interface.

        """
        yield controller.seq.read()
        for mat in controller.bwm:
            yield mat
        yield controller.original

    def debwt_window(self: object) -> None:
        """This method creates a Tkinter Toplevel window for the step-by-step
        inverse BWT protocol, The output file of the protocol will be shown at
        the beginning.

        """
        if self.file:
            debwt_window = Toplevel(self)
            debwt_window.title("Reversing Burros-Wheeler Transform")
            debwt_window.geometry("1000x1000")
            debwt_window.configure(bg='#ebebeb')
            controller = BWDecoder(self.file)
            controller.decode()
            protocol = self.DeBW_output(controller)
            prot= list(protocol)
            
            reconstructed = ["Step 2: Creating the Burros-Wheeler Matrix" for n in range(len(prot) - 2)]

            names = (step for step in ["Step 1 : Visualizing the sequence",
                     *reconstructed,
                     "Step 3: The original sequence is the one that has a $ sign as a last column",
                     "Please refer to the main menu to select another sequence"])

            self.step_by_step(debwt_window, iter(prot), names)
            self.program_output(debwt_window, controller.debwt_output)

        else:
            self.no_file_error()

    def Huff_output(self: object, controller: HuffEncoder) -> Iterator[str]:
        """This method is used to collect all output for Huffman encoding.

        Parameters
        ----------
        controller : HuffEncoder
            The given controller.

        Yields
        ------
        Iterator[str]
            The content to be shown in the interface.

        """
        yield controller.seq.read()
        yield controller.header
        yield controller.binary
        yield controller.unicode
        yield controller.compressed

    def huffcode_window(self: object) -> None:
        """This method creates a Tkinter Toplevel window for the step-by-step
        Huffman coding protocol, The output file of the protocol will be shown
        at the beginning.

        """
        if self.file:
            huff_code_window = Toplevel(self)
            huff_code_window.title("Huffman coding")
            huff_code_window.geometry("1000x1000")
            huff_code_window.configure(bg='#ebebeb')
            names = (step for step in ["Step 1 : Visualizing the sequence",
                     "Step 2: Creating Huffman tree from sequence and calculating paths",
                     "Step 3: Generating the binary sequence from paths and adding a padding",
                     "Step 4: Coding the binary in 8-bits to unicode",
                     "Step 5: Writing paths and unicode to an output file",
                     "Please refer to the main menu to select another sequence"])
            controller = HuffEncoder(self.file)
            controller.encode()
            protocol = self.Huff_output(controller)
            self.step_by_step(huff_code_window, protocol, names)
            self.program_output(huff_code_window, controller.huff_output)

        else:
            self.no_file_error()

    def deHuff_output(self: object, controller: HuffDecoder) -> Iterator[str]:
        """This method is used to collect all output for the Huff decoding.

        Parameters
        ----------
        controller : HuffDecoder
            The given controller.

        Yields
        ------
        Iterator[str]
            The content to be shown in the interface.

        """
        yield controller.seq.read()
        yield controller.header
        yield controller.unicode
        yield controller.binary 
        yield controller.decompressed

    def huffdecode_window(self: object) -> None:
        """This method creates a Tkinter Toplevel window for the step-by-step
        Huffman decoding, The output file of the protocol will be shown at the 
        beginning.

        """
        if self.file:
            huff_decode_window = Toplevel(self)
            huff_decode_window.title("Huffman decoding")
            huff_decode_window.geometry("1000x1000")
            huff_decode_window.configure(bg='#ebebeb')
            names = (step for step in ["Step 1 : Visualizing the compressed sequence",
                     "Step 2: Reading the header of the file that corresponds to Huffman codes that where created with the tree during compression",
                     "Step 3: Separating the unicode sequence",
                     "Step 4: Transforming the unicode sequence to binary using huffman codes in the header and stripping padding",
                     "Step 5: The decompressed sequence : ",
                     "Please refer to the main menu to select another sequence"])
            controller = HuffDecoder(self.file)
            controller.decode()
            protocol = self.deHuff_output(controller)
            self.step_by_step(huff_decode_window, protocol, names)
            self.program_output(huff_decode_window, controller.dehuffman_output)

        else:
            self.no_file_error()   

    def fullzip_output(self: object, controller: FullEncoder) -> Iterator[str]:
        """This method is used to collect all output for a full protocol
        encoding (BWT + Huffman compression).

        Parameters
        ----------
        controller : FullEncoder
            The given controller.

        Yields
        ------
        Iterator[str]
            The content to be shown in the interface.

        """
        yield controller.bw_encoder.seq.read()
        for rot in controller.bw_encoder.rotations:
            yield rot
        yield controller.bw_encoder.bwm
        yield controller.bw_encoder.bwt
        yield controller.huff_encoder.header
        yield controller.huff_encoder.binary
        yield controller.huff_encoder.unicode
        yield controller.huff_encoder.compressed

    def fullzip_window(self: object) -> None:
        """This method creates a Tkinter Toplevel window for the step-by-step
        Fullzip protocol (BWT+Huffman compression), The output file of the 
        protocol will be shown at the beginning.

        """
        if self.file:
            fullzip_window = Toplevel(self)
            fullzip_window.title("Burrow-Wheeler Transform + Huffman coding")
            fullzip_window.geometry("1000x1000")
            fullzip_window.configure(bg='#ebebeb')
            controller = FullEncoder(self.file)
            controller.full_zip()
            protocol = self.fullzip_output(controller)
            prot= list(protocol)

            rots = ["Step 2: Generating all rotations of the sequence" for n in range(len(prot) - 7)]
            names = (step for step in ["Step 1 : Visualizing the sequence",
                     *rots,
                     "Step 3: Creating the Burros-Wheeler matrix by sorting all rotations",
                     "Step 4: The Burros-Wheeler transform is the last column of the matrix",
                     "Step 5: Creating Huffman tree from Burros-Wheeler transform and calculating paths",
                     "Step 6: Generating the binary sequence from paths and adding a padding",
                     "Step 7: Coding the binary in 8-bits to unicode",
                     "Step 8: Writing paths and unicode to an output file",
                     "Please refer to the main menu to select another sequence"])

            self.step_by_step(fullzip_window, iter(prot), names)
            outputs = controller.bw_encoder.bwt_output + \
                '\n' + controller.huff_encoder.huff_output
            self.program_output(fullzip_window, outputs)

        else:
            self.no_file_error()

    def fullunzip_output(self: object, controller: FullDecoder) -> Iterator[str]:
        """This method is used to collect all output for a full decompression
        protocol.

        Parameters
        ----------
        controller : FullDecoder
            The given controller.

        Yields
        ------
        Iterator[str]
            The content to be shown in the interface.

        """
        yield controller.huff_decoder.seq.read()
        yield controller.huff_decoder.header
        yield controller.huff_decoder.unicode
        yield controller.huff_decoder.binary 
        yield controller.huff_decoder.decompressed
        for line in controller.bw_decoder.bwm:
            yield line
        yield controller.bw_decoder.original

    def fullunzip_window(self: object) -> None:
        """This method creates a Tkinter Toplevel window for the step-by-step
        Full unzipping protocol (HUffman decoding + inverse BWT), The output 
        file of the protocol will be shown at the beginning.

        """
        if self.file:
            fullunzip_window = Toplevel(self)
            fullunzip_window.title("Huffman decoding + reverse Burros-Wheeler transform")
            fullunzip_window.geometry("1000x1000")
            fullunzip_window.configure(bg='#ebebeb')
            controller = FullDecoder(self.file)
            controller.full_unzip()
            protocol = self.fullunzip_output(controller)
            prot= list(protocol)

            reconstructed = ["Step 6: Creating the Burros-Wheeler Matrix" for n in range(len(prot) - 6)]
            names = (step for step in ["Step 1 : Visualizing the compressed sequence",
                     "Step 2: Reading the header of the file that corresponds to Huffman codes that where created with the tree during compression",
                     "Step 3: Separating the unicode sequence",
                     "Step 4: Transforming the unicode sequence to binary using huffman codes in the header and stripping padding",
                     "Step 5: The decompressed sequence is the burros wheeler transform of the original sequence:",
                     *reconstructed,
                     "Step 7: The original sequence is the one that has a $ sign as a last column in the Burros-Wheeler Matrix",
                     "Please refer to the main menu to select another sequence"])

            self.step_by_step(fullunzip_window, iter(prot), names)
            outputs = controller.huff_decoder.dehuffman_output + \
                '\n' + controller.bw_decoder.debwt_output
            self.program_output(fullunzip_window, outputs)
        else:
            self.no_file_error()

    def create_menu(self: object) -> None:
        """Creates the menu for the interface."""
        menubar = Menu(self)
        menuFile = Menu(menubar, tearoff=0)
        menubar.add_cascade(label="Menu", menu=menuFile)
        #menubar.add_cascade(label="About", menu=menuFile)
        menuFile.add_command(label="Choose a file", command=self.open_file,
                             accelerator="Ctrl+o")
        menuFile.add_command(label="About", command=self.about)
        self.bind_all("<Control-o>", lambda e: self.open_file())
        self.config(menu=menubar)

    def no_file_error(self: object) -> None:
        """Shows an error when no file is selected."""
        messagebox.showerror("No file selected", "Please select a file")

    def program_output(self: object, window: Tk,  path: str) -> None:
        """Shows the output of a given protocol.

        Parameters
        ----------
        window : Tk
            The given window.
        path : str
            The output file path to be shown in the messagebox.

        """
        messagebox.showinfo(parent=window, title="Program output",
                            message="The output will be saved to \n %s" %(
                                path))

    def main(self: object) -> None:
        """Launches the mainloop of the interface."""
        print("[View] main")
        self.mainloop()

    def quit(self: object) -> None:
        """Quits the current interface."""
        self.destroy()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Encoder classes, Controller architecture in a MVC layout."""
from __future__ import absolute_import
import os
from dnazip.sequence import Sequence
from dnazip.burros_wheeler import BurrosWheeler
from dnazip.huffman import HuffmanTree

class BWEncoder:
    """An encoder class for Burros-Wheeler transform, it is used as a 
    controller in a MVC architecture.

    Attributes
    ----------
    path: str
        The path of the file to be transformed with Burros-Wheeler.
    seq: Sequence
        The sequence that was extracted from the file; a Sequence object.
    bwt_output: str
        The output file path for BWT.
    rotations: List[str]
        A matrix of rotations from the original sequence.
    bwm: List[str]
        The Burros-Wheeler Matrix.
    bwt: str
        The Burros-Wheeler transform of the sequence.
    """
    def __init__(self: object, path: str) -> None:
        """Class constructor.

        Parameters
        ----------
        path : str
            The path of the file to be read.

        Returns
        -------
        None
            A class instance.

        """
        self.path = os.path.splitext(path)[0]
        self.seq = Sequence(path)
        self.bwt_output = self.path + '_bwt.txt'
        self.rotations = None
        self.bwm = None
        self.bwt = None

    def encode(self: object) -> None:
        """The main encoding method of the controller.

        Returns
        -------
        None
            Fills all the properties of an object and writes out the
            transformed sequence to a file.

        """
        self.rotations = list(BurrosWheeler.string_rotations(self.seq.read()))
        self.bwm = BurrosWheeler.construct_bwm(self.rotations[-1])
        self.bwt = BurrosWheeler.encode_bwt(self.bwm)
        Sequence(self.bwt_output).write(self.bwt)

class HuffEncoder:
    """An encoder class for Huffman compression, it is used as a controller
    in a MVC architecture.
    
    Attributes
    ----------
    path: str
        The path of the file to be compressed with Huffman compression.
    seq: Sequence
        The sequence that was extracted from the file; a Sequence object.
    huff_output: str
        The output file path for Huffman compression.
    binary: str
        The binary sequence that was translated from the original sequence 
        using Huffman tree and codes.
    header: str
        The header of the compressed file; contains Huffman codes and paths
        as well as padding that were generated when compressing the sequence.
    unicode: str
        The compressed format of the sequence.
    compressed: str
        The compressed sequence to be written to a file.
    """

    def __init__(self: object, path: str) -> None:
        """Class constructor.

        Parameters
        ----------
        path : str
            The path of the file to be read.

        Returns
        -------
        None
            A class instance.

        """
        self.path = os.path.splitext(path)[0]
        self.seq = Sequence(path)
        self.huff_output = self.path + '_compressed.txt'
        self.binary = None
        self.header = None
        self.unicode = None
        self.compressed = None

    def encode(self: object) -> None:
        """The main encoding method of the controller.

        Returns
        -------
        None
            Fills all the properties of an object and writes out the
            compressed sequence to a file.

        """
        tree = HuffmanTree(self.seq.read())
        tree.get_codings(tree.root)
        self.binary = tree.seq_to_binstr()
        self.unicode = HuffmanTree.binstr_to_unicode(self.binary)
        self.header = tree.codes_to_header()
        self.compressed =  self.header + self.unicode
        Sequence(self.huff_output).write_bytes(self.compressed)

class FullEncoder:
    """An encoder class for both the Burros-Wheeler transform and Huffman
    compression, controller architecture.
    
    Attributes
    ----------
    path: str
        The path of the file to be compressed with BWT + Huffman compression.
    bw_encoder: BWEncoder
        A BWEncoder object to do the Burros-Wheeler transform on a sequence.
    huff_encoder: HuffEncoder
        A HuffEncoder object to do the Huffman compression on the BW transform.
    """

    def __init__(self: object, path: str) -> None:
        """Class constructor.

        Parameters
        ----------
        path : str
            The path of the file to be read.

        Returns
        -------
        None
            A class instance.

        """
        self.path = path
        self.bw_encoder = None
        self.huff_encoder = None

    def full_zip(self: object) -> None:
        """The main encoding method of the controller, it first encodes the
        sequence with BWT, then passes the BWT to Huffman compression.

        Returns
        -------
        None
            Fills all the properties of an object and writes out the
            compressed sequence to a file.

        """
        self.bw_encoder = BWEncoder(self.path)
        self.bw_encoder.encode()
        
        self.huff_encoder = HuffEncoder(self.bw_encoder.bwt_output)
        self.huff_encoder.encode()



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name='dnazip-bioinfo',
    version='0.1',
    description='A Python implementation of the Burros-Wheeler and Huffman coding algorithms',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='http://github.com/dabane-ghassan/dnazip',
    author='dabane-ghassan',
    author_email = 'dabane.ghassan@gmail.com',
    download_url = 'https://github.com/dabane-ghassan/dnazip/archive/v0.1.tar.gz',
    license='MIT',
    packages=setuptools.find_packages(include=['dnazip', 'dnazip.*']),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    #install_requires=REQUIREMENTS,
    extras_require={
        'dev': [
            'pytest >= 6.0.0',
            'pytest-cov >= 2.10.0',
            'coveralls >= 2.1.2',
            'flake8 >= 3.8.0',
            'mock >= 4.0.0',
        ]
    },
    entry_points={
        'gui_scripts': [
            'dnazip=dnazip.main:main'
        ]
    },
    python_requires='>=3.6',
)



# -*- coding: utf-8 -*-
"""
A class to read and write DNA sequences (or a list of characters) to a file.
In a MVC structure, this represents the Model of the data.
"""
from __future__ import absolute_import
import random

class Sequence:
    """A class to represent a sequence, sequences can be read from files or
    writtes to files.

    Attributes
    ----------
    path: str
        The path of the file to be read.

    """

    def __init__(self: object, path: str) -> None:
        """Class constructor.

        Parameters
        ----------
        path : str
            The path of the file to be read.

        Returns
        -------
        None
            A class instance.

        """
        self.path = path

    def __str__(self: object) -> str:
        """This method returns a string representation of the object with
        print().

        Returns
        -------
        str
            A user friendly string representation.

        """
        return "Hey there! I'm a Sequence object"

    def __repr__(self: object) -> str:
        """This method returns a coder-friendly string representation of the
        object.

        Returns
        -------
        str
            A string representation of a sequence object.

        """
        return "Sequence(%s)" % self.path

    def read(self: object) -> str:
        """This method is used to read the contents of the file into a python
        string.

        Returns
        -------
        str
            The contents of the file, i.e. the sequence.

        """
        seq = ""
        with open(self.path, 'r') as file:
            for line in file:
                seq += line.strip("\n")
        return seq

    def read_bytes(self: object) -> str:
        """This method is used to read a file that has been written in bytes,
        it will be useful when reading files that has been compressed with
        Huffman coding.

        Returns
        -------
        str
            The contents of the file, i.e. the sequence.

        """
        seq = ""
        with open(self.path, 'rb') as file:
            for line in file:
                seq += line.decode("utf-8")
        return seq

    def write(self: object, content: str) -> None:
        """This method is used to write out to a file.

        Parameters
        ----------
        content : str
            The contents of the file.

        Returns
        -------
        None
            Writes out to a new file.

        """
        with open(self.path, 'w') as file:
            file.writelines(content)

    def write_bytes(self: object, content: str) -> None:
        """This method is used to write out to a new file into a bytes format,
        UTF-8 encoding, useful when we want to write out the contents of the
        Huffman compression to a file.

        Parameters
        ----------
        content : str
            The contents of the file.

        Returns
        -------
        None
            Writes out to a new file in a bytes format.

        """
        with open(self.path, 'wb') as file:
            file.write(content.encode("utf-8"))

    @staticmethod
    def generate(length: int) -> str:
        """This method is used to generate a random DNA sequence using the
        random python library.

        Returns
        -------
        str
            The random DNA sequence.

        """
        return ''.join([random.choice('ATCGN') for _ in range(0, length, 1)])
